unit LibxmlHelper;

interface

uses
  Classes, libxml2, libxslt, SysUtils, LogHelper;

const
  MAX_LOG_THREADS = 1000;

type
  TXslOutput = (xoHTML, xoXML, xoFO, xoDefault);

  ThreadLog = record
    threadId: Cardinal;
    currentErrors: TStringList;
  end;

  TXslRendering = class
  protected
    logCnt: Integer;
    currentLogs: array [0..MAX_LOG_THREADS-1] of ThreadLog;
    logsLock: TMultiReadExclusiveWriteSynchronizer;
    procedure destroyLogs();
    function getMyThreadLog(): Integer;
    function getErrorMessages(): String; virtual;
    procedure addErrorMessage(error: String); virtual;
    procedure clearErrorMessages(); virtual;

  public

    constructor Create();
    destructor Destroy(); override;

    function GetXslOutput(const XslText: string): TXslOutput;
    function GetXslFile(XslFileName: string): string;

    function ApplyXslFileToXml(
            XmlText: string;
            XslFileName: string;
            xslOutput: TXslOutput = xoDefault;
            params: TStringList = nil): string; overload;

    function ApplyXslFileToXml(
            XmlText: xmlDocPtr;
            XslFileName: string;
            xslOutput: TXslOutput = xoDefault;
            params: TStringList = nil): string; overload;

    function ApplyXslFileToXmlForDOM(
            XmlText: xmlDocPtr;
            XslFileName: string;
            xslOutput: TXslOutput = xoDefault;
            params: TStringList = nil): xmlDocPtr;

    function ApplyXslToXml(
            XmlText: xmlDocPtr;
            XslText: string;
            xslUri: String = '';
            xslOutput: TXslOutput = xoDefault;
            params: TStringList = nil
            ): string; overload;

    function ApplyXslToXml(
            XmlText, XslText: string;
            xslUri: String = '';
            xslOutput: TXslOutput = xoDefault;
            params: TStringList = nil
            ): string; overload;

    function ApplyXslToXml(
            xmlSrcDoc: xmlDocPtr;
            xslSrcDoc: xsltStylesheetPtr;
            params: TStringList = nil
            ): xmlDocPtr; overload;

    function ApplyXslToXmlForDom(
            XmlText: xmlDocPtr;
            XslText: string;
            xslUri: String = '';
            xslOutput: TXslOutput = xoDefault;
            params: TStringList = nil
            ): xmlDocPtr; overload;

    function applyXPath(xmlstr, xpathStr: String): String; overload;
    function applyXPath(doc: xmlDocPtr; xpath: String; nodeSetCopy: Boolean = False): String; overload;

    function parseStyleSheet(xslText: String; url: String = ''): xsltStylesheetPtr;
    function parseDocument(xmlText: String): xmlDocPtr;

    function dumpDocument(doc: xmlDocPtr; output: TXslOutput = xoDefault;
      omitXmlDeclaration: Boolean = False): String;
    function dumpNodeset(nodeset: xmlNodeSetPtr): String;

    procedure dumpXmlOmitDeclaration(out_doc: xmlDocPtr; doc_txt_ptr: PPChar;
      doc_txt_len: PLongInt);
  end;

var
  GlobalXslRendering: TXslRendering;

implementation

uses
  XmlHelper, Windows;

{ TGeoXslRendering }

function format(msg: String; args: array of Cardinal): String;
var
  iarg,i: Integer;
  res: String;
begin
  res := '';
  iarg := 0;
  i := 1;

  while i<=length(msg) do
    begin
    if msg[i] <> '%' then
      res := res + msg[i]
    else
      begin
      i := i+1;

      case msg[i] of
        'd': res := res + IntToStr(Integer(args[iarg]));
        's': res := res + StrPas(PChar(args[iarg]));
        'c': res := res + Char(args[iarg]);
        else
          begin
          iArg := iArg-1;
          res := res + msg[i];
          end;
      end;
      iArg := iArg+1;
      end;
    i := i+1;
    end;
    Result := res;
end;


procedure errorHandler(
          ctx: pointer;
          msg: PChar;
          arg1,arg2,arg3,arg4: Cardinal
          ); cdecl;
var
 ret: String;
begin
 if (msg = nil) or (ctx = nil) then exit;
 ret := format(adjustLinebreaks(StrPas(msg)),[arg1, arg2, arg3,arg4]);
 TXslRendering(ctx).addErrorMessage(ret);
end;

constructor TXslRendering.Create();
begin
   self.logsLock := TMultiReadExclusiveWriteSynchronizer.Create();
   self.logCnt := 0;
   xmlInitParser();
   xmlSetGenericErrorFunc(self,xmlGenericErrorFunc(@errorHandler));
   xsltSetGenericErrorFunc(self,xmlGenericErrorFunc(@errorHandler));
//   exsltCommonRegister(); // mc2.6 is this needed???
//   exsltDynRegister();
end;

destructor TXslRendering.Destroy();
begin
   inherited;
   xsltCleanupGlobals();
   self.destroyLogs();
   self.logsLock.free();
end;

function TXslRendering.ApplyXslFileToXml(XmlText: string;
  XslFileName: string; xslOutput: TXslOutput;params: TStringList): string;
var
  xslstr, filename: string;
  XslStream: TFileStream;
begin
  filename := XslFileName;
  if not FileExists(filename) then
    raise Exception.Create('ApplyXslFileToXml: XSL file "' + filename + '" not found');

  // carica l'XSL da uno stream
  XslStream := TFileStream.Create(filename,
    fmOpenRead + fmShareDenyWrite);
  try
    SetLength(xslstr, XslStream.size);
    XslStream.Read(xslstr[1], XslStream.size);
    // TODO: redefine encoding
    xslstr := CheckXmlEncoding(xslstr);
    // do the actual conversion
    try
      // URI must use '/' separators
      Result := ApplyXslToXml(XmlText, xslstr,
        FixPathDelimiters (xslFileName, '/'), xslOutput, params);
    except
      on E: Exception do
      begin
        raise Exception.Create('ApplyXslFileToXml: ' + E.Message + ' in file ' + filename);
      end
    end;
    // Result := AdjustLineBreaks(Result); // TODO: determinare se serve
  finally
    XslStream.Free;
  end;
end;

function TXslRendering.ApplyXslFileToXml(XmlText: xmlDocPtr;
  XslFileName: string; xslOutput: TXslOutput;params: TStringList): string;
var
  xslstr, filename: string;
  XslStream: TFileStream;
begin
  filename := XslFileName;
  if not FileExists(filename) then
    raise Exception.Create('ApplyXslFileToXml: XSL file "' + filename + '" not found');

  // carica l'XSL da uno stream
  XslStream := TFileStream.Create(filename,
    fmOpenRead + fmShareDenyWrite);
  try
    SetLength(xslstr, XslStream.size);
    XslStream.Read(xslstr[1], XslStream.size);
    // TODO: redefine encoding
    xslstr := CheckXmlEncoding(xslstr);
    // do the actual conversion
    try
      // URI must use '/' separators
      Result := ApplyXslToXml(XmlText, xslstr,
        FixPathDelimiters (xslFileName, '/'), xslOutput, params);
    except
      on E: Exception do
      begin
        raise Exception.Create('ApplyXslFileToXml: ' + E.Message + ' in file ' + filename);
      end
    end;
    Result := AdjustLineBreaks(Result); // TODO: determinare se serve
  finally
    XslStream.Free;
  end;
end;


function TXslRendering.ApplyXslFileToXmlForDOM(XmlText: xmlDocPtr;
  XslFileName: string; xslOutput: TXslOutput;
  params: TStringList): xmlDocPtr;
var
  xslstr, filename: string;
  XslStream: TFileStream;
begin
  Result := nil;
  filename := XslFileName;
  if not FileExists(filename) then
    raise Exception.Create('ApplyXslFileToXml: XSL file "' + filename + '" not found');

  // carica l'XSL da uno stream
  XslStream := TFileStream.Create(filename,
    fmOpenRead + fmShareDenyWrite);
  try
    SetLength(xslstr, XslStream.size);
    XslStream.Read(xslstr[1], XslStream.size);
    // TODO: redefine encoding
    xslstr := CheckXmlEncoding(xslstr);
    // do the actual conversion
    try
      // URI must use '/' separators
      Result := ApplyXslToXmlForDom(XmlText, xslstr,
        FixPathDelimiters (xslFileName, '/'), xslOutput, params);
    except
      on E: Exception do
      begin
        raise Exception.Create('ApplyXslFileToXml: ' + E.Message + ' in file ' + filename);
      end
    end;
  finally
    XslStream.Free;
  end;
end;

function TXslRendering.GetXslFile(XslFileName: string): string;
var
  XslStream: TFileStream;
begin
  if not FileExists(xslfilename) then
    raise Exception.Create('ApplyXslFileToXml: XSL file "' + xslfilename + '" not found');

  // carica l'XSL da uno stream
  XslStream := TFileStream.Create(xslfilename,
    fmOpenRead + fmShareDenyWrite);
  try
    SetLength(Result, XslStream.size);
    XslStream.Read(Result[1], XslStream.size);
    Result := AdjustLineBreaks(Result);
  finally
    XslStream.Free;
  end;
end;


function TXslRendering.ApplyXslToXml(
            XmlText, XslText: string;
            xslUri: String;
            xslOutput: TXslOutput;
            params: TStringList
            ): string;
var
  xslSrcDoc: xsltStylesheetPtr;
  xmlSrcDoc,xmlResDoc : xmlDocPtr;
begin
//  xmlText := CheckXmlEncoding(xmlText);
  xmlSrcDoc := nil;
  xslSrcDoc := nil;
  xmlResDoc := nil;

  try
      xslSrcDoc := parseStyleSheet(xslText, xslUri);
      xmlSrcDoc := parseDocument(xmlText);

      xmlResDoc := applyXslToXml(xmlSrcDoc, xslSrcDoc, params);
      Result := dumpDocument(xmlResDoc,xslOutput);
  finally
      if Assigned(xmlSrcDoc) then xmlFreeDoc(xmlSrcDoc);
      if Assigned(xmlResDoc) then xmlFreeDoc(xmlResDoc);
      if Assigned(xslSrcDoc) then xsltFreeStylesheet(xslSrcDoc);
  end;
end;

function TXslRendering.ApplyXslToXml(
            xmlSrcDoc: xmlDocPtr;
            xslSrcDoc: xsltStylesheetPtr;
            params: TStringList
            ): xmlDocPtr;
var
  tmpP,FXSLTParams : PPChar;

  xmlResDoc : xmlDocPtr;
  tc  : xsltTransformContextPtr;
  i: Integer;
begin

  tc := nil;
  Result := nil;
  FXSLTParams := nil;

  if params <> nil then
    begin
    FXSLTParams := xmlMalloc((params.Count*2+1)* sizeOf(PChar));
    tmpP := FXSLTParams;
    for i:=0 to params.Count-1 do
      begin
      tmpP^ := xmlStrdup(PChar(params.Names[i]));
      inc(tmpP, 1);
      tmpP^ := xmlStrdup(PChar(''''+ params.Values[params.Names[i]] + ''''));
      inc(tmpP, 1);
      end;
    tmpP^ := nil;
    end;

  try

      xmlSubstituteEntitiesDefault(1);
      if (xslSrcDoc = nil) then
        raise Exception.Create('XSL file is null.');

      if (xmlSrcDoc = nil) then
        raise Exception.Create('XML file is null.');

      tc := xsltNewTransformContext(xslSrcDoc, xmlSrcDoc);

      xmlResDoc := xsltApplyStylesheetUser(xslSrcDoc, xmlSrcDoc, PPChar(FXSLTParams), 'xml' , nil, tc);

      if xmlResDoc = nil then
        begin
          raise Exception.Create('XSL Error: ' + self.getErrorMessages());
        end;

      Result := xmlResDoc;
  finally
    if FXSLTParams <> nil then
      xmlFree(PChar(FXSLTParams));

      if Assigned(tc) then xsltFreeTransformContext(tc);
      //xsltCleanupGlobals();
      self.clearErrorMessages();
  end;
end;

function TXslRendering.ApplyXslToXmlForDom(XmlText: xmlDocPtr;
  XslText: string; xslUri: String; xslOutput: TXslOutput;
  params: TStringList): xmlDocPtr;
var
  xslSrcDoc: xsltStylesheetPtr;
begin
  xslSrcDoc := nil;
  try
      xslSrcDoc := parseStyleSheet(xslText, xslUri);
      Result := applyXslToXml(XmlText, xslSrcDoc, params);
  finally
      if Assigned(xslSrcDoc) then xsltFreeStylesheet(xslSrcDoc);
      //xsltCleanupGlobals();
  end;
end;

function TXslRendering.applyXPath(doc: xmlDocPtr; xpath: String; nodeSetCopy: Boolean): String;
var
  xpathCtx: xmlXPathContextPtr;
  xpathResult: xmlXPathObjectPtr;
  tmpRes: PChar;
begin
   tmpRes := nil;
   xpathCtx := nil;
   xpathResult := nil;

   Result := '';

   try
     xpathCtx := xmlXPathNewContext(doc);
     xpathResult := xmlXPathEval(PChar(xpath),xpathCtx);
     if xpathResult = nil then
        Exception.Create('XPath error: '  + getErrorMessages());
     if nodeSetCopy then
       result := dumpNodeSet(xpathResult^.nodesetval)
     else
       begin
       tmpRes := xmlXPathCastToString(xpathResult);
       Result := StrPas(tmpRes);
       end;
   finally
     if (tmpRes <> nil) then xmlFree(tmpRes);
     if assigned(xpathCtx) then xmlXPathFreeContext(xpathCtx);
     if assigned(xpathResult) then xmlXPathFreeObject(xpathResult);
     self.clearErrorMessages();
   end;

end;

function TXslRendering.dumpDocument(doc: xmlDocPtr;
  output: TXslOutput = xoDefault; omitXmlDeclaration: Boolean = False): String;
const
  chtml = 'html';
  cxml  = 'xml';
var
  Len: Longint;
  buffer,outputMethod: PChar;
begin
  case output of
    xoHTML:
      begin
      outputMethod := chtml;
      end;
    xoXML:
      begin
      outputMethod := cxml;
      end;
    xoFO:
      begin
      outputMethod := cxml;
      end;
    xoDefault:
      begin
      outputMethod := chtml;
      end
    else outputMethod := chtml;
  end;

  try
      if outputMethod = chtml then
          htmlDocDumpMemory(doc, @Buffer, @Len)
      else begin
        if not(omitXmlDeclaration) then
          xmlDocDumpMemory(doc, @Buffer, @Len)  // mc2.6 (added@)
        else
          self.dumpXmlOmitDeclaration(doc, @Buffer, @Len);
      end;

      Result := StrPas(Buffer);
  finally
    if Assigned(Buffer) then xmlFree(Buffer);
  end;
end;

function TXslRendering.parseDocument(xmlText: String): xmlDocPtr;
begin
  try
    Result := xmlParseDoc(pchar(xmlText));
    if (Result = nil) then
      begin
        raise Exception.Create('XML document is not wellformed: ' + self.getErrorMessages());
      end;
  finally
    self.clearErrorMessages();
  end;

end;


function TXslRendering.parseStyleSheet(xslText: String; url: String): xsltStylesheetPtr;
var
  doc: xmlDocPtr;
begin
  Result := nil;
  try
    doc := xmlParseDoc(pchar(xslText));
    if (doc = nil) then
      begin
        raise Exception.Create('XML document is not wellformed: ' + self.getErrorMessages());
      end;

    doc^.URL := xmlStrdup(PChar(url));
    Result := xsltParseStylesheetDoc(doc);
    if (Result = nil) then
      raise Exception.Create('XSL file is not a valid xsl: ' + getErrorMessages());
  finally
    self.clearErrorMessages();
  end;

end;


function TXslRendering.GetXslOutput(
  const XslText: string): TXslOutput;
begin
  if Pos('<fo:root', XslText) > 0 then
    Result := xoFO
  else
    Result := xoHTML;
end;

procedure TXslRendering.clearErrorMessages();
var
  i: Cardinal;
begin
  i := self.getMyThreadLog();

  if self.currentLogs[i].currentErrors.Count > 0 then
    LogError('XSL WARN:' + self.currentLogs[i].currentErrors.text);

  self.currentLogs[i].currentErrors.clear();
end;

function TXslRendering.getErrorMessages(): String;
var
  i: Cardinal;
begin
  i := self.getMyThreadLog();
  Result := self.currentLogs[i].currentErrors.Text;
end;

procedure TXslRendering.addErrorMessage(error: String);
var
  i: Cardinal;
begin
  i := self.getMyThreadLog();
  self.currentLogs[i].currentErrors.Add(error);
end;

function TXslRendering.ApplyXslToXml(XmlText: xmlDocPtr;
  XslText: string; xslUri: String;
  xslOutput: TXslOutput;
  params: TStringList): string;
var
  xslSrcDoc: xsltStylesheetPtr;
  xmlResDoc : xmlDocPtr;
begin
  xslSrcDoc := nil;
  xmlResDoc := nil;

  try
      xslSrcDoc := parseStyleSheet(xslText, xslUri);
      xmlResDoc := applyXslToXml(XmlText, xslSrcDoc, params);
      Result := dumpDocument(xmlResDoc,xslOutput);
  finally
      if Assigned(xmlResDoc) then xmlFreeDoc(xmlResDoc);
      if Assigned(xslSrcDoc) then xsltFreeStylesheet(xslSrcDoc);
      //xsltCleanupGlobals();
  end;
end;

function TXslRendering.applyXPath(xmlstr, xpathStr: String): String;
var
  doc: xmlDocPtr;
begin
  doc := GlobalXslRendering.parseDocument(xmlstr);
  try
    Result := GlobalXslRendering.applyXPath(doc, xpathStr);
  finally
    xmlFreeDoc(doc);
  end;
end;


function TXslRendering.dumpNodeset(nodeset: xmlNodeSetPtr): String;
var
   buffer: xmlBufferPtr;
begin
   Result := '';
   if ((nodeset = nil) or (nodeset^.nodeNr = 0) or (nodeset^.nodeTab = nil)) then
      exit;

   buffer := xmlBufferCreate();
   try
   xmlNodeDump(buffer,nodeset^.nodeTab^.doc,nodeset^.nodeTab^,0,0);
   result := StrPas(buffer^.content);
   finally
     xmlBufferFree(buffer);
   end;

end;

function TXslRendering.getMyThreadLog(): Integer;
var
  i: Integer;
  myThreadId: Cardinal;
begin
  myThreadId := getCurrentThreadId(); // windows only
  i := 0;
  self.logsLock.BeginRead();
  try
    while (i<self.logCnt) and (self.currentLogs[i].threadId<>myThreadId) do
      inc(i);
  finally
    self.logsLock.EndRead();
  end;
  if  (i=self.logCnt) then
    begin
    self.logsLock.BeginWrite();
    try
      inc(self.logCnt);
      i := self.logCnt-1;
      self.currentLogs[i].threadId := myThreadId;
      self.currentLogs[i].currentErrors := TStringList.Create();
    finally
      self.logsLock.EndWrite();
    end;
    end;

  Result := i;
end;

procedure TXslRendering.destroyLogs();
var
  i: Integer;
begin
  i := 0;
  while (i<self.logCnt) do
    begin
    FreeAndNil(self.currentLogs[i].currentErrors);
    inc(i);
    end;
end;

procedure TXslRendering.dumpXmlOmitDeclaration(out_doc: xmlDocPtr;
  doc_txt_ptr: PPChar; doc_txt_len: PLongInt);
var
    doc_charset: xmlCharEncoding;
    out_buff: xmlOutputBufferPtr;
    conv_hdlr: xmlCharEncodingHandlerPtr;
    child: xmlNodePtr;
    txt_encoding: PChar;
begin
    if (doc_txt_ptr = nil) then begin
        raise Exception.Create(
           'xmlDocDumpFormatMemoryEnc:  Null return buffer pointer.'
           );
    end;

    conv_hdlr := nil;
    child := nil;
    doc_txt_ptr^ := nil;
    doc_txt_len^ := 0;

    if (out_doc = nil) then begin
        raise Exception.Create(
           'xmlDocDumpFormatMemoryEnc:  Null DOM tree document pointer.'
           );
    end;

    (*
     *  Validate the encoding value, if provided.
     *  This logic is copied from xmlSaveFileEnc.
     *)

    txt_encoding := out_doc^.encoding;
    if (txt_encoding <> nil) then begin
        doc_charset := xmlParseCharEncoding(txt_encoding);

        if (xmlCharEncoding(out_doc^.charset) <> XML_CHAR_ENCODING_UTF8) then begin
            raise Exception.Create(
               'xmlDocDumpFormatMemoryEnc: Source document not in UTF8'
               );
        end else if (doc_charset <> XML_CHAR_ENCODING_UTF8) then begin
            conv_hdlr := xmlFindCharEncodingHandler(txt_encoding);
            if ( conv_hdlr <> nil ) then begin
                   raise Exception.Create(
                     'xmlDocDumpFormatMemoryEnc:' +
                     'Failed to identify encoding handler for ' +
                     'character set ' + txt_encoding
                     );
            end;
        end;
    end;
    out_buff := xmlAllocOutputBuffer(conv_hdlr);
    if (out_buff = nil ) then begin
        raise Exception.Create(
            'xmlDocDumpFormatMemoryEnc: Failed to allocate output buffer.'
          );
    end;

    if (out_doc^.children <> nil) then
        child := out_doc^.children;

    while (child <> nil) do begin
        xmlNodeDumpOutput(out_buff, out_doc, child, 0, 0, txt_encoding);
        xmlOutputBufferWriteString(out_buff, sLineBreak);
        child := child^.next;
      end;

    xmlOutputBufferFlush(out_buff);
    if (out_buff^.conv <> nil) then begin
        doc_txt_len^ := out_buff^.conv^.use;
        doc_txt_ptr^ := xmlStrndup(out_buff^.conv^.content, doc_txt_len^);
    end else begin
      doc_txt_len^ := out_buff^.buffer^.use;
        doc_txt_ptr^ := xmlStrndup(out_buff^.buffer^.content, doc_txt_len^);
    end;
    xmlOutputBufferClose(out_buff);

    if ((doc_txt_ptr^ = nil) and (doc_txt_len^ > 0)) then begin
        raise Exception.Create(
           'xmlDocDumpFormatMemoryEnc: ' +
           'Failed to allocate memory for document text representation.'
           );
    end;

end;

initialization
  GlobalXslRendering := TXslRendering.Create;
finalization
  GlobalXslRendering.Free;
end.

